import 'package:plexus/expose.dart';
import 'package:plexus/plexus.dart';

/// A set of [Localization] for [supportedLanguages]
class Localizations extends UnmodifiableMapView<Locale, Localization> {
  /// Creates a new instance of [Localizations]
  const Localizations(this.fallbackLocale, this.supportedLanguages, this.supportedLocales, Map<Locale, Localization> localizations) : super(localizations);

  /// The locales supported by these localizations
  final Set<Locale> supportedLocales;

  /// The languages supported by these localizations
  final Set<Language> supportedLanguages;

  /// The [Locale] used if an unsupported language is used to access these localizations.
  final Locale fallbackLocale;

  /// Checks whether [locale] is supported by these localizations
  bool isLocaleSupported(Locale locale) => supportedLocales.contains(locale);

  /// Checks whether [language] is supported by these localizations
  bool isLanguageSupported(Language language) => supportedLanguages.contains(language);

  /// Prefer this to the indexing operator, which should only be used for locales that are guaranteed to exist.
  Localization operator |(LanguageBase language) => language is Locale ? (isLocaleSupported(language) ? internalMap[language] : isLanguageSupported(language.language)) : isLanguageSupported(language) ? internalMap[language] : internalMap[fallbackLocale];
}

/// A set of localized messages/formats/etc. for [locale]
class Localization extends UnmodifiableMapView<String, dynamic> {
  /// Creates a new instance of [Localization]
  const Localization(this.locale, Map<String, dynamic> localizations) : super(localizations);

  /// The locale of this [Localization]
  final Locale locale;

  /// Prefer this to the indexing operator, which should only be used for keys that are guaranteed to exist.
  dynamic operator |(String key) => internalMap.containsKey(key) ? internalMap[key] : key;
}

/// A representation of a locale consisting of a language and, optionally, a country.
class Locale extends LanguageBase {
  const Locale._(this.language, [this.country]) : super._();

  /// The language of this locale
  final Language language;

  /// The country of this locale
  @nullable
  final Country country;

  @override
  String get code => language.code;

  @override
  String get label => language.label;

  /// Common English
  static const common_english = Locale._(Language.english);

  /// Canadian English
  static const canadian_english = Locale._(Language.english, Country.canada);
  
  /// British English
  static const british_english = Locale._(Language.english, Country.united_kingdom);
  
  /// American English
  static const american_english = Locale._(Language.english, Country.united_states);

  /// Chinese
  static const chinese = Locale._(Language.chinese, Country.china);

  /// Common French
  static const common_french = Locale._(Language.french);

  /// Canadian French
  static const canadian_french = Locale._(Language.french, Country.canada);

  /// French French
  static const french_french = Locale._(Language.french, Country.france);
  
  /// Spanish Spanish
  static const spanish_spanish = Locale._(Language.spanish, Country.spain);

  /// Locales supported by Plexus
  static const locales = {
    common_english,
    canadian_english,
    british_english,
    american_english,
    chinese,
    common_french,
    canadian_french,
    french_french,
    spanish_spanish,
  };
}

/// A template for a representation of a language or locale
abstract class LanguageBase {
  const LanguageBase._();

  /// The ISO 639-1 representation of this language
  String get code;

  /// The name of this language (in English)
  String get label;
}

/// A representation of a language
class Language extends LanguageBase {
  const Language._(this.code, this.label) : super._();

  @override
  final String code;

  @override
  final String label;

  /// The Chinese language
  static const chinese = Language._('zh', 'Chinese');

  /// The Czech language
  static const czech = Language._('cs', 'Czech');

  /// The Dutch language
  static const dutch = Language._('nl', 'Dutch');
  
  /// The English language
  static const english = Language._('en', 'English');
  
  /// The French language
  static const french = Language._('fr', 'French');
  
  /// The German language
  static const german = Language._('de', 'German');
  
  /// The Spanish language
  static const spanish = Language._('es', 'Spanish');
  
  /// The Welsh language
  static const welsh = Language._('cy', 'Welsh');

  /// Countries supported by Plexus
  static const languages = {
    chinese,
    czech,
    dutch,
    english,
    french,
    german,
    spanish,
    welsh
  };
}

/// A representation of a country
class Country {
  const Country._(this.code, this.label);

  /// The ISO 3166-1 representation of this country.
  final String code;

  /// The name of this country (in English).
  final String label;

  /// Australia
  static const australia = Country._('AU', 'Australia');

  /// Canada
  static const canada = Country._('CA', 'Canada');

  /// China
  static const china = Country._('CN', 'China');
  
  /// France
  static const france = Country._('FR', 'France');
  
  /// Germany
  static const germany = Country._('DE', 'Germany');

  /// India
  static const india = Country._('IN', 'India');

  /// Spain
  static const spain = Country._('ES', 'Spain');
  
  /// The United Kingdom
  static const united_kingdom = Country._('UK', 'United Kingdom');

  /// The United States of America
  static const united_states = Country._('US', 'United States of America');

  /// Countries supported by Plexus
  static const all = {
    australia,
    canada,
    china,
    france,
    germany,
    india,
    spain,
    united_kingdom,
    united_states,
  };
}
