import 'dart:async';
import 'dart:io';

import 'package:intl/intl.dart';

/// A DateFormat implementing ISO 8601 without timezone
final iso8601NoTZ = DateFormat('yyyy-MM-ddTHH-mm-ss.SSS');

/// The user's home directory
Directory get userHome => Directory(Platform.environment['HOME'] ?? Platform.environment['USERPROFILE'] ?? '/sdcard');

/// Can be used to bypass a parse function.
T loopback<T>(T t) => t;

/// Prints [value] using the native [print] method, optionally appending and prepending newlines to [value].
void printSpaced(dynamic value, {bool leadingNewline = false, bool trailingNewline = true}) => print('${leadingNewline ? '\n' : ''}$value${trailingNewline ? '\n' : ''}');

/// A [RegExp] pattern for all horizontal whitespace characters,
final whitespacePattern = RegExp(r'\p{Blank}');

/// When used to annotate a field/parameter, that field/parameter is able to store a null value (prior to NNBD)
class Nullable {
  const Nullable._();
}

/// When used to annotate a field/parameter, that field/parameter is able to store a null value (prior to NNBD)
const Nullable nullable = Nullable._();

/// A template for an optionally-asynchronous callback that returns a particular [R] when given a particular [V1]
typedef FutureOr<R> FutureOrValueCallback<V1, R>(V1 value1);

/// A template for an optionally-asynchronous callback that returns a particular [R] when given a particular [V1] and [V2]
typedef FutureOr<R> FutureOrValue2Callback<V1, V2, R>(V1 value1, V2 value2);

/// A template for an optionally-asynchronous callback that returns a particular [R] when given a particular [V1], [V2] and [V3]
typedef FutureOr<R> FutureOrValue3Callback<V1, V2, V3, R>(V1 value1, V2 value2, V3 value3);

/// A template for an optionally-asynchronous standalone callback that returns a particular [R]
/// 
/// Typically used for zero-input generators
typedef FutureOr<R> FutureOrCallback<R>();

/// A template for an optionally-asynchronous callback that tests a value of type [V]
typedef FutureOr<bool> FutureOrValueTest<V>(V value);


/// A template for a callback that returns a particular [R] when given a particular [V]
typedef R ValueCallback<V1, R>(V1 value1);

/// A template for a callback that returns a particular [R] when given a particular [V1] and [V2]
typedef R Value2Callback<V1, V2, R>(V1 value1, V2 value2);

/// A template for a callback that returns a particular [R] when given a particular [V1], [V2] and [V3]
typedef R Value3Callback<V1, V2, V3, R>(V1 value1, V2 value2, V3 value3);

/// A template for a standalone callback that returns a particular [R]
/// 
/// Typically used for zero-input generators
typedef R Callback<R>();

/// A template for a callback that tests a value of type [V]
typedef bool ValueTest<V>(V value);

/// Placeholder event handler
void voidEventHandler(_) {}

/// Placeholder asynchronous event handler
FutureOr<void> asyncVoidEventHandler(_) {}

/// Placeholder callback
void voidHandler() {}

/// Placeholder asynchronous callback
Future<void> asyncVoidHandler() async {}

/// Base class for basic operator wrappers
abstract class OperatorWrapper {}

/// Implements indexing behaviour in the form of retrieval by, setting by, and removal by keys 
class IndexingOperatorWrapper<K, V> implements OperatorWrapper {
  /// Creates a new instance of [IndexingOperatorWrapper]
  const IndexingOperatorWrapper(this.onGet, this.onSet, this.onRemove);

  /// Called to retrieve a value by key.
  final ValueCallback<K, V> onGet;

  /// Called to set a value by key.
  final Value2Callback<K, V, void> onSet;

  /// Called to remove a value by key.
  final ValueCallback<K, void> onRemove;

  /// Returns the value of `key`.
  V operator [](K key) => onGet(key);

  /// Sets the value of `key` to `value`.
  void operator []=(K key, V value) => onSet(key, value);

  /// Removes `key` and its value
  void operator |(K key) => onRemove(key);
}

/// Implements [List] indexing behaviour in the form of retrieval by, setting by, and removal by [int] keys
class ListOperatorWrapper<E> extends IndexingOperatorWrapper<int, E> {
  /// Creates a new instance of [ListOperatorWrapper]
  const ListOperatorWrapper(ValueCallback<int, E> onGet, Value2Callback<int, E, void> onSet, ValueCallback<int, void> onRemove) : super(onGet, onSet, onRemove);
}
